unit XClasses;

//{$DEFINE USE_SYSUTILS}
// Using of sysutils is increasing program size from 16,5K to 36,5K
// because of using Exception class. Refusing from it makes applet
// smaller, but if item could not be added to ZList, this does not
// lead to stop further execution.

interface
   uses Windows {$IFDEF USE_SYSUTILS}, SysUtils {$ENDIF};
   // Windows is needed because we are programming for Windows not for DOS...

const
   MaxListSize = Maxint div 16;

type
   // ZList is intended to exchange VCL's TList.
   // It does not support sorting but using of it instead of
   // standard TList allow program to be reduced by 20K
   // because of no needs in Classes unit.

   APointerList = ^ZPointerList;
   ZPointerList = array[0..MaxListSize - 1] of Pointer;

   {$IFDEF USE_SYSUTILS}
   EZListError = class( Exception );
   {$ENDIF}

   ZList = class( TObject )
   private
      fItems : APointerList;
      fCount : Integer;
      fCapacity : Integer;
   protected
      procedure SetCapacity( Value : Integer );
      function Get( Idx : Integer ) : Pointer;
      procedure Put( Idx : Integer; Value : Pointer );
   public
      destructor Destroy; override;
      procedure Dispose;
      procedure Clear;
      function Add( Value : Pointer ) : Integer;
      procedure Delete( Idx : Integer );
      function IndexOf( Value : Pointer ) : Integer;
      {$IFDEF USE_SYSUTILS}
      class procedure Error(const Msg: string); dynamic;
      {$ENDIF}
      property Count : Integer read fCount;
      property Capacity : Integer read fCapacity write SetCapacity;
      property Items[ Idx : Integer ] : Pointer read Get write Put;
   end;

   XOnEvent = procedure( Sender : TObject ) of object;

   // Main difference of XClass from its ancestor TObject is a
   // property Parent which is corresponding to property Owner
   // of appropriate VCL base object known as TComponent.
   // Parent of XClass also is of XClass type and it is responsible
   // for destroying of its childs when destroying.
   XClass = class( TObject )
   private
      fParent : XClass;
      fChilds : ZList;
      procedure SetParent( Value : XClass );
      function GetChildsCount : Integer;
      function GetChilds( Idx : Integer ) : XClass;
   protected
   public
      constructor Create( AParent : XClass ); dynamic;
      destructor Destroy; override;
      property Parent : XClass read fParent write SetParent;
      property ChildsCount : Integer read GetChildsCount;
      property Childs[ Idx : Integer ] : XClass read GetChilds;
      function AddChild( Child : XClass ) : Integer; dynamic;
      procedure DeleteChild( Child : XClass ); dynamic;
   end;

   XChild = class( XClass )
   private
   protected
   public
      function AddChild( Child : XClass ) : Integer; override;
   end;

   function XRect( Left, Top, Right, Bottom : Integer ) : TRect;
   function XPoint( X, Y : Integer ) : TPoint;

implementation

function XRect( Left, Top, Right, Bottom : Integer ) : TRect;
begin
   Result.Left := Left;
   Result.Top  := Top;
   Result.Right:= Right;
   Result.Bottom := Bottom;
end;

function XPoint( X, Y : Integer ) : TPoint;
begin
   Result.x := X;
   Result.y := Y;
end;

{ ZList implementation }

destructor ZList.Destroy;
begin
   Clear;
   inherited;
end;

{$IFDEF USE_SYSUTILS}
class procedure ZList.Error(const Msg: string);

  function ReturnAddr: Pointer;
  asm
          MOV     EAX,[EBP+4]
  end;

begin
  raise EZListError.Create(Msg) at ReturnAddr;
end;
{$ENDIF}

procedure ZList.SetCapacity( Value : Integer );
var NewItems : APointerList;
begin
   if Value < Count then
      Value := Count;
   if Value = fCapacity then Exit;
   GetMem( NewItems, Sizeof( Pointer ) * Value );
   if fItems <> nil then
   begin
      if Count > 0 then
         Move( fItems[ 0 ], NewItems[ 0 ], Count * Sizeof( Pointer ) );
      FreeMem( fItems );
   end;
   fItems := NewItems;
   fCapacity := Value;
end;

procedure ZList.Dispose;
var I : Integer;
begin
   for I := 0 to Count - 1 do
       FreeMem( fItems[ I ] );
   Clear;
end;

procedure ZList.Clear;
begin
   if fItems <> nil then
      FreeMem( fItems );
   fItems := nil;
   fCount := 0;
   fCapacity := 0;
end;

function ZList.Add( Value : Pointer ) : Integer;
begin
   Result := -1;
   if Count >= Capacity then
      Capacity := Count + 1;
   if Count >= Capacity then
      {$IFDEF USE_SYSUTILS}
      Error( 'Can not add value to a list' );
      {$ELSE}
      Exit;
      {$ENDIF}
   fItems[ fCount ] := Value;
   Result := fCount;
   fCount := fCount + 1;
end;

procedure ZList.Delete( Idx : Integer );
begin
   if Idx < 0 then Exit;
   if Idx >= Count then Exit;
   Move( fItems[ Idx + 1 ], fItems[ Idx ], Sizeof( Pointer ) * (Count - Idx - 1) );
   Dec( fCount );
end;

procedure ZList.Put( Idx : Integer; Value : Pointer );
begin
   if Idx < 0 then Exit;
   if Idx >= Count then Exit;
   fItems[ Idx ] := Value;
end;

function ZList.Get( Idx : Integer ) : Pointer;
begin
   Result := nil;
   if Idx < 0 then Exit;
   if Idx >= Count then Exit;
   Result := fItems[ Idx ];
end;

function ZList.IndexOf( Value : Pointer ) : Integer;
var I : Integer;
begin
   Result := -1;
   for I := 0 to Count - 1 do
   begin
      if fItems[ I ] = Value then
      begin
         Result := I;
         break;
      end;
   end;
end;

{ XClass implementation }

constructor XClass.Create( AParent : XClass );
begin
   inherited Create;
   fChilds := ZList.Create;
   Parent := AParent;
end;

destructor XClass.Destroy;
var I : Integer;
    X : XClass;
begin
   for I := fChilds.Count - 1 downto 0 do
   begin
      X := fChilds.Items[ I ];
      X.Free;
   end;
   fChilds.Free;
   if fParent <> nil then
   begin
      I := fParent.fChilds.IndexOf( Self );
      if I >= 0 then
         fParent.fChilds.Delete( I );
   end;
   inherited;
end;

procedure XClass.SetParent( Value : XClass );
begin
   if Value = fParent then Exit;
   if fParent <> nil then
      fParent.DeleteChild( Self );
   fParent := Value;
   if fParent <> nil then
      fParent.AddChild( Self );
end;

function XClass.GetChildsCount : Integer;
begin
   Result := fChilds.Count;
end;

function XClass.GetChilds( Idx : Integer ) : XClass;
begin
   Result := nil;
   if (Idx < 0) or (Idx >= ChildsCount) then
      Exit;
   Result := fChilds.Items[ Idx ];
end;

function XClass.AddChild( Child : XClass ) : Integer;
begin
   Result := fChilds.Add( Child );
end;

procedure XClass.DeleteChild( Child : XClass );
var I : Integer;
begin
   I := fChilds.IndexOf( Child );
   if I >= -1 then
      fChilds.Delete( I );
end;

{ XChild }

function XChild.AddChild(Child: XClass) : Integer;
begin
   Result := -1;
end;

end.
